
From: Bjorn Helgaas <bjorn.helgaas@hp.com>

This is based on 2.6.9-rc1-mm2, i.e., it applies after bk-acpi.patch,
bk-input.patch, and fix-smm-failures-on-e750x-systems.patch from -mm2.

Add ACPI-based i8042 keyboard and aux controller enumeration.

This can be disabled with "i8042.no_acpi=1".  If you need to disable it,
please let me know so I can fix it.

The main reason we need this is because i8042_check_aux() does a
request_irq(12) before it knows whether the AUX controller is even present.
 Some boards with a keyboard controller but no AUX controller reuse IRQ 12
for PCI interrupts.  And apparently the BIOS on some of those boards leaves
IRQ 12 programmed as edge-triggered, active high, even though the PCI
interrupt is level-triggered, active low.  So if i8042 comes along and
requests IRQ 12 before the PCI driver enables the device and programs the
IOAPIC correctly, i8042 gets an endless stream of interrupts.

This patch avoids that situation by only poking at ports that ACPI tells us
about.  If ACPI doesn't tell us about an AUX controller, we don't touch IRQ
12.

Signed-off-by: Bjorn Helgaas <bjorn.helgaas@hp.com>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 25-akpm/Documentation/kernel-parameters.txt |    1 
 25-akpm/drivers/input/serio/i8042.c         |  150 ++++++++++++++++++++++++++++
 2 files changed, 151 insertions(+)

diff -puN Documentation/kernel-parameters.txt~acpi-based-i8042-keyboard-aux-controller-enumeration Documentation/kernel-parameters.txt
--- 25/Documentation/kernel-parameters.txt~acpi-based-i8042-keyboard-aux-controller-enumeration	2004-09-02 21:04:48.582936576 -0700
+++ 25-akpm/Documentation/kernel-parameters.txt	2004-09-02 21:04:48.590935360 -0700
@@ -483,6 +483,7 @@ running once the system is up.
 			     controller
 	i8042.reset	[HW] Reset the controller during init and cleanup
 	i8042.unlock	[HW] Unlock (ignore) the keylock
+	i8042.no_acpi=1	[HW] Don't use ACPI to discover KBD/AUX ports
 
 	i810=		[HW,DRM]
 
diff -puN drivers/input/serio/i8042.c~acpi-based-i8042-keyboard-aux-controller-enumeration drivers/input/serio/i8042.c
--- 25/drivers/input/serio/i8042.c~acpi-based-i8042-keyboard-aux-controller-enumeration	2004-09-02 21:04:48.584936272 -0700
+++ 25-akpm/drivers/input/serio/i8042.c	2004-09-02 21:04:48.591935208 -0700
@@ -24,6 +24,15 @@
 #include <linux/pci.h>
 #include <linux/err.h>
 
+#ifdef CONFIG_ACPI
+#include <linux/acpi.h>
+#include <acpi/acpi_bus.h>
+
+static int no_acpi;
+module_param(no_acpi, uint, 0);
+MODULE_PARM_DESC(no_acpi, "Disable ACPI keyboard/aux controller detection");
+#endif
+
 #include <asm/io.h>
 
 MODULE_AUTHOR("Vojtech Pavlik <vojtech@suse.cz>");
@@ -1067,6 +1076,142 @@ static struct serio * __init i8042_alloc
 	return serio;
 }
 
+#ifdef CONFIG_ACPI
+static int acpi_kbd_registered;
+static int acpi_aux_registered;
+
+struct i8042_resources {
+	unsigned int port1;
+	unsigned int port2;
+	unsigned int irq;
+};
+
+static acpi_status acpi_i8042_resource(struct acpi_resource *res, void *data)
+{
+	struct i8042_resources *i8042 = data;
+	struct acpi_resource_io *io;
+	struct acpi_resource_irq *irq;
+	struct acpi_resource_ext_irq *ext_irq;
+
+	if (res->id == ACPI_RSTYPE_IO) {
+		io = &res->data.io;
+		if (io->range_length) {
+			if (!i8042->port1)
+				i8042->port1 = io->min_base_address;
+			else
+				i8042->port2 = io->min_base_address;
+		}
+	} else if (res->id == ACPI_RSTYPE_IRQ) {
+		irq = &res->data.irq;
+		if (irq->number_of_interrupts > 0)
+			i8042->irq = acpi_register_gsi(irq->interrupts[0],
+				irq->edge_level, irq->active_high_low);
+	} else if (res->id == ACPI_RSTYPE_EXT_IRQ) {
+		ext_irq = &res->data.extended_irq;
+		if (ext_irq->number_of_interrupts > 0)
+			i8042->irq = acpi_register_gsi(ext_irq->interrupts[0],
+				ext_irq->edge_level, ext_irq->active_high_low);
+	}
+	return AE_OK;
+}
+
+static int acpi_i8042_kbd_add(struct acpi_device *device)
+{
+	struct i8042_resources i8042;
+	acpi_status status;
+
+	memset(&i8042, 0, sizeof(i8042));
+	status = acpi_walk_resources(device->handle, METHOD_NAME__CRS,
+		acpi_i8042_resource, &i8042);
+	if (ACPI_FAILURE(status))
+		return -ENODEV;
+
+	printk("i8042: ACPI %s [%s] at I/O 0x%x, 0x%x, irq %d\n",
+		acpi_device_name(device), acpi_device_bid(device),
+		i8042.port1, i8042.port2, i8042.irq);
+
+	i8042_data_reg = i8042.port1;
+	i8042_command_reg = i8042.port2;
+	i8042_status_reg = i8042.port2;
+	i8042_kbd_values.irq = i8042.irq;
+
+	return 0;
+}
+
+static int acpi_i8042_aux_add(struct acpi_device *device)
+{
+	struct i8042_resources i8042;
+	acpi_status status;
+
+	memset(&i8042, 0, sizeof(i8042));
+	status = acpi_walk_resources(device->handle, METHOD_NAME__CRS,
+		acpi_i8042_resource, &i8042);
+	if (ACPI_FAILURE(status))
+		return -ENODEV;
+
+	printk("i8042: ACPI %s [%s] at irq %d\n",
+		acpi_device_name(device), acpi_device_bid(device), i8042.irq);
+
+	i8042_aux_values.irq = i8042.irq;
+
+	return 0;
+}
+
+static struct acpi_driver acpi_i8042_kbd_driver = {
+	.name		= "i8042",
+	.ids		= "PNP0303",
+	.ops		= {
+		.add		= acpi_i8042_kbd_add,
+	},
+};
+
+static struct acpi_driver acpi_i8042_aux_driver = {
+	.name		= "i8042",
+	.ids		= "PNP0F13",
+	.ops		= {
+		.add		= acpi_i8042_aux_add,
+	},
+};
+
+static int acpi_i8042_init(void)
+{
+	int err;
+
+	if (no_acpi) {
+		printk("i8042: ACPI detection disabled\n");
+		return -ENODEV;
+	}
+
+	err = acpi_bus_register_driver(&acpi_i8042_kbd_driver);
+	if (err >= 0)
+		acpi_kbd_registered = 1;
+	if (err == 0)
+		return 0;
+
+	err = acpi_bus_register_driver(&acpi_i8042_aux_driver);
+	if (err >= 0)
+		acpi_aux_registered = 1;
+	if (err == 0)
+		i8042_noaux = 1;
+	return 1;
+}
+
+static void acpi_i8042_exit(void)
+{
+	if (acpi_kbd_registered) {
+		acpi_bus_unregister_driver(&acpi_i8042_kbd_driver);
+		acpi_kbd_registered = 0;
+	}
+	if (acpi_aux_registered) {
+		acpi_bus_unregister_driver(&acpi_i8042_aux_driver);
+		acpi_aux_registered = 0;
+	}
+}
+#else
+static inline int  acpi_i8042_init(void) { return -ENODEV; }
+static inline void acpi_i8042_exit(void) { }
+#endif
+
 int __init i8042_init(void)
 {
 	int i;
@@ -1083,6 +1228,9 @@ int __init i8042_init(void)
 	i8042_aux_values.irq = I8042_AUX_IRQ;
 	i8042_kbd_values.irq = I8042_KBD_IRQ;
 
+	if (acpi_i8042_init() == 0)
+		return -ENODEV;
+
 	if (i8042_controller_init())
 		return -ENODEV;
 
@@ -1146,6 +1294,8 @@ void __exit i8042_exit(void)
 
 	del_timer_sync(&i8042_timer);
 
+	acpi_i8042_exit();
+
 	platform_device_unregister(i8042_platform_device);
 	driver_unregister(&i8042_driver);
 
_
